import os
import re
import six

try:
    from os import scandir
except ImportError:
    from scandir import scandir

import OpenSSL

from .certificate import CertificateChain
from .exceptions import CertificateExpiredException, CertificateScanFailedException, DuplicateCertificateException, \
    DuplicatePrivateKeyException


PATTERN_CERTIFICATE_HEADER = r'-----BEGIN CERTIFICATE-----'
PATTERN_PRIVATE_KEY_HEADER = r'-----BEGIN (RSA )?PRIVATE KEY-----'


def scan_for_certificates(path='.'):
    """
    Scan for certificates in the specified path

    :param str path: Path to directory containing certificates
    :return: Dictionary of certificate paths
    :rtype: dict
    """
    certs = []
    keys = []

    # Scan the specified directory
    for entry in scandir(path=path):
        # Only open files
        if not os.path.isfile(entry.path):
            continue

        # Read the first line of each file and look for a regex
        abs_path = os.path.abspath(entry.path)
        with open(abs_path, 'r') as f:
            line = f.readline()

            if re.match(PATTERN_CERTIFICATE_HEADER, line):
                # Seek to the start of the file and open the certificate
                f.seek(0)
                cert_data = f.read()
                cert = OpenSSL.crypto.load_certificate(OpenSSL.crypto.FILETYPE_PEM, cert_data.encode('utf8'))

                # Check if this certificate already exists
                for item in certs:
                    if item['cert'].get_subject() == cert.get_subject():
                        raise DuplicateCertificateException(
                            'ERROR: Certificate "%s" duplicates "%s"' % (abs_path, item['path'])
                        )

                # Add to the list
                certs.append({'cert': cert, 'path': abs_path})

            elif re.match(PATTERN_PRIVATE_KEY_HEADER, line):
                # Seek to the start of the file and open the key
                f.seek(0)
                private_key_data = f.read()
                private_key = OpenSSL.crypto.load_privatekey(OpenSSL.crypto.FILETYPE_PEM, private_key_data)

                # Check if this key already exists
                for item in keys:
                    private_numbers = item['key'].to_cryptography_key().private_numbers()
                    if private_numbers == private_key.to_cryptography_key().private_numbers():
                        raise DuplicatePrivateKeyException(
                            'ERROR: Private key "%s" duplicates "%s"' % (abs_path, item['path'])
                        )

                # Add to the list
                keys.append({'key': private_key, 'path': abs_path})

    # Make sure a certificate and private key was found
    if len(certs) == 0:
        raise CertificateScanFailedException('ERROR: No certificates were found')
    elif len(certs) == 1:
        raise CertificateScanFailedException('ERROR: Only one certificate was found. The chain may be missing.')
    if len(keys) == 0:
        raise CertificateScanFailedException('ERROR: No private keys were found')

    # Parse the certificates
    cert_info = parse_certificates(certs, keys)

    # Perform validation
    if not validate_certificates(cert_info):
        raise CertificateScanFailedException(
            'ERROR: Unable to automatically find the correct certificates. Please enter the certificate manually '
            'using the command line arguments.'
        )

    # Transform the objects to their paths
    for key, value in cert_info.items():
        if isinstance(value, dict) and 'path' in value:
            cert_info[key] = value['path']
        elif isinstance(value, CertificateChain):
            cert_info[key] = value.path

    return cert_info


def validate_certificates(cert_info):
    """
    Validate that the provided certificates have not expired and are correct

    :param dict cert_info: Dictionary containing certificate, chain, and private key information
    :return: Boolean
    :rtype: bool
    """
    # Ensure none of the certificates have expired
    if cert_info['certificate'].cert.has_expired():
        raise CertificateExpiredException('ERROR: Base certificate "%s" has expired.' % cert_info['certificate'].path)
    elif cert_info['chain'].cert.has_expired():
        raise CertificateExpiredException('ERROR: Issuer certificate "%s" has expired.' % cert_info['chain'].path)

    # Print out the certificate details for verification by the user
    print('Certificate:')
    print('Path = %s' % cert_info['certificate'].path)
    for (key, value) in cert_info['certificate'].cert.get_subject().get_components():
        print('%s = %s' % (key.decode(), value.decode()))
    print('')

    # Print out the chain details for verification by the user
    print('Issuer Certificate:')
    print('Path = %s' % cert_info['chain'].path)
    for (key, value) in cert_info['chain'].cert.get_subject().get_components():
        print('%s = %s' % (key.decode(), value.decode()))
    print('')

    # Print out the private key path for verification by the user
    print('Private Key:')
    print('Path = %s' % cert_info['private_key']['path'])
    print('')

    # Request verification from the user
    if six.PY2:
        input_method = raw_input
    else:
        input_method = input

    return input_method('Are these files correct? (y/n): ') == 'y'


def parse_certificates(certs, keys):
    """
    Parse the certificates that have been scanned and find the base certificate, corresponding private key, and issuer.

    :param list of dict certs: List of dictionary objects formatted as {'cert': X509 object, 'path': 'file/path.crt'}
    :param list of dict keys: List of dictionary objects formatted as {'key': PKey object, 'path': 'file/path.pem'}
    :return: Dictionary of information about the certificate, chain, and private key
    :rtype: dict
    """
    # Build the certificate chain
    first_cert = None
    cert_chain = []
    for cert in certs:
        # Create the chain item
        chain_item = CertificateChain(cert['cert'], cert['path'])

        if not first_cert:
            # Save the first certificate
            first_cert = chain_item
        else:
            # Get the subject and issuer for this certificate
            subject = cert['cert'].get_subject()
            issuer = cert['cert'].get_issuer()

            # Evaluate relationships for all items in the chain
            for item in cert_chain:
                # Go to the top
                current = item
                while current.parent:
                    current = current.parent

                # Check for parents/children at the top
                if current.cert.get_issuer() == subject:
                    # chain_item is a parent
                    current.parent = chain_item
                    chain_item.child = current
                elif current.cert.get_subject() == issuer:
                    # chain_item is a child
                    current.child = chain_item
                    chain_item.parent = current

                # Go to the bottom
                while current.child:
                    current = current.child

                # Check for parents/children at the bottom
                if current.cert.get_issuer() == subject:
                    # chain_item is a parent
                    current.parent = chain_item
                    chain_item.child = current
                elif current.cert.get_subject() == issuer:
                    # chain_item is a child
                    current.child = chain_item
                    chain_item.parent = current

        # Add to the chain list
        cert_chain.append(chain_item)

    # Fail out if more than one base certificate was found
    base_certs = [item for item in cert_chain if not item.child]
    if len(base_certs) > 1:
        raise CertificateScanFailedException(
            'ERROR: More than one base certificate was found: %s' % ', '.join([cert.path for cert in base_certs])
        )

    # Save the base cert
    base_cert = base_certs[0]
    del base_certs

    # Get the public key of the base cert
    public_key = base_cert.cert.get_pubkey()

    # Find the private key that goes with this public key
    private_key = None
    for key in keys:
        public_numbers = key['key'].to_cryptography_key().public_key().public_numbers()

        # If the public numbers match, then this is the expected public key
        if public_key.to_cryptography_key().public_numbers() == public_numbers:
            private_key = key
            break

    # If a corresponding private key was not found, raise an error
    if not private_key:
        raise CertificateScanFailedException(
            'ERROR: Could not find corresponding private key for the base certificate "%s".' % base_cert.path
        )

    return {
        'certificate': base_cert,
        'private_key': private_key,
        'chain': base_cert.parent
    }



from pprint import pprint

from .base import CertificateUploader


class ACMCertificateUploader(CertificateUploader):
    """
    Certificate uploader

    Uploads new certificates and applies them to an ELB
    """

    def __init__(self, profile=None, role=None):
        super(ACMCertificateUploader, self).__init__(profile, role)
        self._acm_client = self.session.client('acm')

    def get_server_certificate(self, arn):
        """
        Get a certificate from ACM

        :param str arn: Certificate ARN
        :return: Certificate information
        :rtype: dict
        """
        cert = self._acm_client.describe_certificate(CertificateArn=arn)
        return cert['Certificate']

    def upload_certificate(self, cert_data, private_key_data, chain_data, dry_run=False):
        """
        Upload an IAM Server Certificate

        :param str cert_data: Path to the certificate file on the local machine
        :param str private_key_data: Path to the private key file on the local machine
        :param str chain_data: Path to the certificate chain file on the local machine
        :param bool dry_run: Whether to perform a dry run of the operations. Defaults to False.
        :return: Certificate ARN
        :rtype: str
        """
        if not dry_run:
            response = self._acm_client.import_certificate(
                Certificate=cert_data,
                PrivateKey=private_key_data,
                CertificateChain=chain_data
            )

            # Save the ARN
            arn = response['CertificateArn']

            # Get details about the certificate
            certificate = self.get_server_certificate(arn)

            # Print details about the upload
            print('Certificate Uploaded Successfully:')
            print('\tARN: %s' % arn)
            print('\tCommon Name: %s' % certificate['DomainName'])
            if certificate['SubjectAlternativeNames']:
                print('\tSubject Alternative Names:')
                for item in certificate['SubjectAlternativeNames']:
                    print('\t\t%s' % item)
            print('\tValidity: %s to %s' % (certificate['NotBefore'], certificate['NotAfter']))
            print('\tStatus: %s' % certificate['Status'])
            pprint(certificate)

            return arn
        else:
            print('%sWould upload certificate to ACM:' % '[DRY RUN] ' if dry_run else '')
            print('\tCertificate File Path (local system): %s' % cert_data)
            print('\tPrivate Key File Path (local system): %s' % private_key_data)
            print('\tCertificate Chain File Path (local system): %s' % chain_data)

    def tag_certificate(self, arn, tags=None):
        """
        Add tags to a certificate

        :param str arn: Certificate ARN
        :param dict tags: Dictionary of tags formatted as follows:
                    {
                        'abc': '123',
                        'def': '456'
                    }
        :return: Response
        """
        if not tags:
            tags = {}

        # Build the tag list
        tag_list = [{'Key': key, 'Value': value} for key, value in tags.items()]

        # Perform the call
        return self._acm_client.add_tags_to_certificate(
            CertificateArn=arn,
            Tags=tag_list
        )



from argparse import ArgumentParser
from time import sleep

from .scan import scan_for_certificates
from .uploader import ACMCertificateUploader, IAMCertificateUploader


def main():
    parser = ArgumentParser(
        description='Upload ACM/IAM certificates and apply them to ELBs',
        usage=
        '''
        Scan for certificates in the current directory and upload to IAM:
            cert-uploader \\
                --scan \\
                iam \\
                --certificate-name cert-name
        
        Upload a new certificate to IAM:
            cert-uploader \\
                --certificate-path path/to/certificate.crt \\
                --private-key-path path/to/key.pem \\
                --certificate-chain-path path/to/chain.crt \\
                iam
                --certificate-name cert-name
        
        Upload a new certificate to ACM:
            cert-uploader \\
                --certificate-path path/to/certificate.crt \\
                --private-key-path path/to/key.pem \\
                --certificate-chain-path path/to/chain.crt \\
                acm
                
        Upload a new certificate to ACM and add tags:
            cert-uploader \\
                --certificate-path path/to/certificate.crt \\
                --private-key-path path/to/key.pem \\
                --certificate-chain-path path/to/chain.crt \\
                acm \\
                --tag Name=cert \\
                --tag App=app1
                
        Upload a new certificate to IAM and assign it to a load balancer:
            cert-uploader \\
                --certificate-path path/to/certificate.crt \\
                --private-key-path path/to/key.pem \\
                --certificate-chain-path path/to/chain.crt \\
                --load-balancer load-balancer-name \\
                iam \\
                --certificate-name cert-name
                
        Assign an existing IAM certificate to a load balancer:
            cert-uploader \\
                --load-balancer load-balancer-name \\
                iam \\
                --certificate-name cert-name
                
        Assign an existing ACM certificate to a load balancer:
            cert-uploader \\
                --load-balancer load-balancer-name \\
                acm \\
                --certificate-arn arn:aws:acm:REGION:ACCOUNT:certificate/CERTIFICATE_ID
                
        Upload a new IAM certificate at the path "/test" and assign it to a load balancer:
            cert-uploader \\
                --load-balancer load-balancer-name \\
                --certificate-path path/to/certificate.crt \\
                --private-key-path path/to/key.pem \\
                --certificate-chain-path path/to/chain.crt \\
                iam \\
                --certificate-name cert-name \\
                --iam-path /test
                
        Upload a new ACM certificate and assign it to a load balancer on port 8443:
            cert-uploader \\
                --load-balancer load-balancer-name \\
                --port 8443 \\
                --certificate-path path/to/certificate.crt \\
                --private-key-path path/to/key.pem \\
                --certificate-chain-path path/to/chain.crt \\
                acm
        '''
    )
    parser.set_defaults(type=None)
    subparsers = parser.add_subparsers()

    """
    Global Arguments
    """
    parser.add_argument(
        '--certificate-path',
        '-c',
        help='Path to certificate file'
    )

    parser.add_argument(
        '--private-key-path',
        '-k',
        help='Path to private key file'
    )

    parser.add_argument(
        '--certificate-chain-path',
        '-x',
        help='Path to the certificate chain'
    )

    parser.add_argument(
        '--load-balancer',
        '-l',
        help='Load balancer name'
    )

    parser.add_argument(
        '--port',
        '-P',
        help='Port of the listener on the ELB to assign this certificate to. Defaults to 443.',
        type=int,
        default=443
    )

    parser.add_argument(
        '--dry-run',
        '-d',
        help='Perform a dry run of the operations',
        action='store_true'
    )

    parser.add_argument(
        '--scan',
        '-s',
        help='Scan the current directory for certificates',
        action='store_true'
    )

    parser.add_argument(
        '--profile',
        '-p',
        help='AWS profile to use',
        default=None
    )

    parser.add_argument(
        '--role',
        '-r',
        help='AWS role to assume',
        default=None
    )

    """
    AWS Certificate Manager Arguments
    """
    acm_parser = subparsers.add_parser('acm', help='Upload the certificate to AWS Certificate Manager')
    acm_parser.set_defaults(type='acm')

    acm_parser.add_argument(
        '--certificate-arn',
        '-a',
        help='ARN of the certificate',
    )

    acm_parser.add_argument(
        '--tag',
        '-t',
        help='Add tags to the certificate',
        action='append'
    )

    """
    IAM Server Certificate Arguments
    """
    iam_parser = subparsers.add_parser('iam', help='Upload the certificate as an IAM Server Certificate')
    iam_parser.set_defaults(type='iam')

    iam_parser.add_argument(
        '--certificate-name',
        '-n',
        help='Unique name the certificate will be saved as in IAM',
    )

    iam_parser.add_argument(
        '--iam-path',
        '-i',
        help='Path in IAM in which the certificate will be stored. Defaults to "/"',
        default='/'
    )

    # Parse arguments
    options = parser.parse_args()

    # Create the uploader
    if options.type == 'iam':
        uploader = IAMCertificateUploader(profile=options.profile, role=options.role)
    elif options.type == 'acm':
        uploader = ACMCertificateUploader(profile=options.profile, role=options.role)
    else:
        parser.print_help()
        return

    # Upload certificate only if the corresponding files are passed in the arguments
    arn = None
    is_new_cert = False
    cert_info = {}

    # Get the certificate path information
    if options.scan:
        cert_info = scan_for_certificates()
    elif options.certificate_path and options.certificate_chain_path and options.private_key_path:
        cert_info = {
            'certificate': options.certificate_path,
            'private_key': options.private_key_path,
            'chain': options.certificate_chain_path
        }

    if cert_info:
        is_new_cert = True

        # Read the data
        cert_data, private_key_data, chain_data = uploader.read_certificate_files(
            cert_path=cert_info['certificate'],
            private_key_path=cert_info['private_key'],
            chain_path=cert_info['chain']
        )

        # Build arguments
        upload_kwargs = {
            'cert_data': cert_data,
            'private_key_data': private_key_data,
            'chain_data': chain_data,
            'dry_run': options.dry_run
        }

        if options.type == 'iam':
            upload_kwargs.update({'name': options.certificate_name})

        # Perform the upload
        arn = uploader.upload_certificate(**upload_kwargs)

    else:
        # Fetch an existing certificate
        if options.type == 'acm' and options.certificate_arn:
            arn = options.certificate_arn
        elif options.type == 'iam' and options.certificate_name:
            certificate = uploader.get_server_certificate(options.certificate_name)
            arn = certificate.arn

    # Add tags to the certificate
    if options.type == 'acm' and options.tag:

        # Make sure an ARN was specified
        if not arn:
            print('Certificate ARN is not defined. Either upload a new certificate or specify an existing ARN.')
            exit(1)

        # Build the tags
        tag_dict = {}
        for item in options.tag:
            tag_parts = item.split('=')
            tag_dict.update({tag_parts[0]: tag_parts[1]})

        # Tag the certificate
        uploader.tag_certificate(arn=arn, tags=tag_dict)

    # Assign certificate to a load balancer
    if options.load_balancer:

        # Ensure an ARN has been found
        if not arn:
            print('ERROR: Certificate could not be found. Either upload a new certificate or specify an existing one.')
            exit(1)

        # Wait a couple seconds for the certificate to be ready
        if is_new_cert and not options.dry_run:
            for i in range(10, 0, -1):
                print('Waiting for certificate to propagate... %d' %i)
                sleep(1)

            # Add some spacing
            print()

        # Apply the certificate to the ELB
        uploader.assign_certificate(
            lb_name=options.load_balancer,
            lb_port=options.port,
            arn=arn,
            dry_run=options.dry_run
        )


if __name__ == '__main__':
    main()



from datetime import datetime

import boto3
import six
from botocore.exceptions import ClientError

from ..exceptions import InvalidProtocolException, ListenerNotFoundException, LoadBalancerNotFoundException, \
    MultipleCertificatesException


class CertificateUploader(object):
    """
    Certificate uploader

    Uploads new certificates and applies them to an ELB
    """

    def __init__(self, profile=None, role=None):
        self.profile = profile
        self.role = role

        # Assume the role if it is specified
        if self.role:
            timestamp = int(datetime.now().timestamp() * 1000000)
            sts = boto3.client('sts')
            response = sts.assume_role(
                RoleArn=self.role,
                RoleSessionName=str(timestamp),
                DurationSeconds=900
            )

            # Create the session with the role's credentials
            self.session = boto3.Session(
                aws_access_key_id=response['Credentials']['AccessKeyId'],
                aws_secret_access_key=response['Credentials']['SecretAccessKey'],
                aws_session_token=response['Credentials']['SessionToken']
            )
        else:
            self.session = boto3.Session(profile_name=profile)

        self._elb_client = self.session.client('elb')
        self._elbv2_client = self.session.client('elbv2')

    def get_server_certificate(self, identifier):
        """
        Get a certificate

        :param str identifier: Certificate identifier
        :return: Certificate information
        :rtype: dict
        """
        raise NotImplementedError

    @staticmethod
    def read_certificate_files(cert_path=None, private_key_path=None, chain_path=None):
        """
        Read data from certificate files.

        :param str cert_path: Path to the certificate file on the local machine
        :param str private_key_path: Path to the private key file on the local machine
        :param str chain_path: Path to the certificate chain file on the local machine
        :return: Tuple of (certificate data, private key data, chain data)
        :rtype: tuple of str
        """

        cert_data = None
        private_key_data = None
        chain_data = None

        if cert_path:
            # Read certificate body
            with open(cert_path, 'r') as f:
                cert_data = f.read()

        if private_key_path:
            # Read private key
            with open(private_key_path, 'r') as f:
                private_key_data = f.read()

        if chain_path:
            # Read certificate chain
            with open(chain_path, 'r') as f:
                chain_data = f.read()

        return cert_data, private_key_data, chain_data

    def upload_certificate(self, cert_data, private_key_data, chain_data, dry_run=False):
        """
        Upload an IAM Server Certificate

        :param str cert_data: Path to the certificate file on the local machine
        :param str private_key_data: Path to the private key file on the local machine
        :param str chain_data: Path to the certificate chain file on the local machine
        :param bool dry_run: Whether to perform a dry run of the operations. Defaults to False.
        :return: Certificate ARN
        :rtype: str
        """
        raise NotImplementedError

    def _get_alb_listeners(self, arn):
        """
        Get all listeners for a particular ALB load balancer

        :param str arn: Load balancer ARN
        :return: List of listeners
        :rtype: list of dict
        """
        listeners = list()
        marker = None

        while True:
            # Fetch the data
            if marker:
                listener_descriptions = self._elbv2_client.describe_listeners(
                    LoadBalancerArn=arn,
                    Marker=marker
                )
            else:
                listener_descriptions = self._elbv2_client.describe_listeners(
                    LoadBalancerArn=arn
                )

            # Add the listeners to the list
            listeners += listener_descriptions['Listeners']

            # Check for the next listener marker
            if listener_descriptions.get('NextMarker'):
                marker = listener_descriptions['NextMarker']
            else:
                break

        return listeners

    def get_load_balancer(self, name):
        """
        Get ELB/ALB information for the name load balancer

        :param name: Load balancer name
        :return: (Load balancer information, Boolean whether the load balancer is an ALB)
        :rtype: (dict, bool)
        :raises: LoadBalancerNotFoundException
        """
        is_alb = False

        try:
            elbs = self._elb_client.describe_load_balancers(LoadBalancerNames=[name])
            lb_descriptions = elbs['LoadBalancerDescriptions']
            lb = lb_descriptions[0]
        except (KeyError, IndexError, ClientError) as e:
            # If this is a ClientError, only catch LoadBalancerNotFound exceptions
            if isinstance(e, ClientError):
                if e.response['Error']['Code'] != 'LoadBalancerNotFound':
                    raise

            # If an ELB couldn't be found, look for an ALB
            try:
                albs = self._elbv2_client.describe_load_balancers(Names=[name])
                lb_descriptions = albs['LoadBalancers']
                lb = lb_descriptions[0]
                is_alb = True
            except (KeyError, IndexError, ClientError):
                # If this is a ClientError, only catch LoadBalancerNotFound exceptions
                if isinstance(e, ClientError):
                    if e.response['Error']['Code'] != 'LoadBalancerNotFound':
                        raise

                raise LoadBalancerNotFoundException('Load balancer %s could not be found' % name)

        return lb, is_alb

    def assign_certificate(self, lb_name, arn, lb_port=443, dry_run=False):
        """
        Assign a certificate to an ELB

        :param str lb_name: ELB name
        :param str arn: Certificate ARN
        :param int lb_port: Load balancer listener port that the certificate will be assigned to
        :param bool dry_run: Whether to perform a dry run of the operations. Defaults to False.
        :return: Response
        :rtype: dict
        """
        # Check that this is a valid ELB
        lb, is_alb = self.get_load_balancer(lb_name)

        # Validate the ALB listener
        listener = None
        if is_alb:
            # Get listeners
            listeners = self._get_alb_listeners(lb['LoadBalancerArn'])

            # Find the desired listener using the port
            for item in listeners:
                if item['Port'] == lb_port:
                    listener = item
                    break

            # Throw error if listener could not be found
            if not listener:
                raise ListenerNotFoundException('Could not find a listener for port %d' % lb_port)

            # Validate the protocol is HTTPS
            if listener['Protocol'] != 'HTTPS':
                raise InvalidProtocolException('Port %s is not a HTTPS listener' % lb_port)

            # Check if there is more than one certificate on the listener
            certs = listener.get('Certificates', [])
            if len(certs) > 1:
                raise MultipleCertificatesException(
                    'There is more than one certificate on the desired listener. This operation is not '
                    'currently supported by cert-uploader and must be performed manually.'
                )

            # Save the existing certificate
            existing_cert = certs[0]['CertificateArn']
        else:
            # Look up the existing certificate
            listener = None
            for item in lb['ListenerDescriptions']:
                if item['Listener']['LoadBalancerPort'] == lb_port:
                    listener = item['Listener']
                    break

            # Throw error if listener could not be found
            if not listener:
                raise ListenerNotFoundException('Could not find a listener for port %d' % lb_port)

            # Validate the protocol is HTTPS
            if listener['Protocol'] != 'HTTPS':
                raise InvalidProtocolException('Port %s is not a HTTPS listener' % lb_port)

            # Save the existing certificate
            existing_cert = listener['SSLCertificateId']

        if not dry_run:
            print('Replacing certificate %(existing_cert_arn)s with %(new_cert_arn)s to load balancer %(lb_name)s '
                  '(%(lb_dns)s) on port %(lb_port)d.' %
                  {
                      'existing_cert_arn': existing_cert,
                      'new_cert_arn': arn,
                      'lb_name': lb_name,
                      'lb_dns': lb['DNSName'],
                      'lb_port': lb_port
                  })

            # Request verification from the user
            if six.PY2:
                input_method = raw_input
            else:
                input_method = input

            if not is_alb:
                # Assign the certificate to the ELB
                response = self._elb_client.set_load_balancer_listener_ssl_certificate(
                    LoadBalancerName=lb_name,
                    LoadBalancerPort=lb_port,
                    SSLCertificateId=arn
                )

                # Check for a successful response
                if response['ResponseMetadata']['HTTPStatusCode'] == 200:
                    print('Certificate %s applied successfully to ELB %s on port %d' % (arn, lb_name, lb_port))
                else:
                    raise Exception('Unable to apply certificate to ELB: %s' % response)

                # Check for rollback
                if input_method('Would you like to roll back? (y/n): ') == 'y':
                    # Roll back the certificate
                    response = self._elb_client.set_load_balancer_listener_ssl_certificate(
                        LoadBalancerName=lb_name,
                        LoadBalancerPort=lb_port,
                        SSLCertificateId=existing_cert
                    )

                    # Check for a successful response
                    if response['ResponseMetadata']['HTTPStatusCode'] == 200:
                        print('Rolled back to certificate %s on ELB %s port %d' % (existing_cert, lb_name, lb_port))
                    else:
                        raise Exception('Unable to apply certificate to ELB: %s' % response)

            else:
                # Update the certificate
                certs = [{'CertificateArn': arn}]

                # Modify the listener on the ALB
                self._elbv2_client.modify_listener(
                    ListenerArn=listener['ListenerArn'],
                    Certificates=certs
                )
                print('Certificate %s applied successfully to ALB %s on port %d' % (arn, lb_name, lb_port))

                # Check for rollback
                if input_method('Would you like to keep this change? (y/n): ') == 'n':
                    # Update the certificate
                    certs = [{'CertificateArn': existing_cert}]

                    # Modify the listener on the ALB
                    self._elbv2_client.modify_listener(
                        ListenerArn=listener['ListenerArn'],
                        Certificates=certs
                    )
                    print('Rolled back to certificate %s on ALB %s port %d' % (existing_cert, lb_name, lb_port))

        else:
            print('[DRY RUN] Certificate %(existing_cert_arn)s would be replaced with %(new_cert_arn)s on '
                  'load balancer %(lb_name)s (%(lb_dns)s) on port %(lb_port)d.' %
                  {
                      'existing_cert_arn': existing_cert,
                      'new_cert_arn': arn,
                      'lb_name': lb_name,
                      'lb_dns': lb['DNSName'],
                      'lb_port': lb_port
                  })

    def tag_certificate(self, arn, tags=None):
        """Add tags to a certificate"""
        raise NotImplementedError



from .base import CertificateUploader
from ..certificate import Certificate
from ..exceptions import CertificateExistsException


class IAMCertificateUploader(CertificateUploader):
    """
    Certificate uploader

    Uploads new certificates and applies them to an ELB
    """

    def __init__(self, profile=None, role=None):
        super(IAMCertificateUploader, self).__init__(profile, role)
        self._iam = self.session.resource('iam')
        self._iam_client = self.session.client('iam')

    def get_server_certificate(self, name):
        """
        Create a Certificate object from the metadata of an IAM Server Certificate
        :param str name: IAM Server Certificate name
        :return: Certificate object
        :rtype: Certificate
        """
        cert = self._iam.ServerCertificate(name)
        cert.load()
        return Certificate.from_metadata(cert.server_certificate_metadata)

    def upload_certificate(self, cert_data, private_key_data, chain_data, name='', iam_path='/', dry_run=False):
        """
        Upload an IAM Server Certificate
        :param str cert_data: Path to the certificate file on the local machine
        :param str private_key_data: Path to the private key file on the local machine
        :param str chain_data: Path to the certificate chain file on the local machine
        :param str name: Certificate name
        :param str iam_path: Path to save the certificate within IAM. Defaults to "/"
        :param bool dry_run: Whether to perform a dry run of the operations. Defaults to False.
        :return: Certificate object
        :rtype: Certificate
        """
        # Check if a certificate by this name already exists
        cert = self._iam.ServerCertificate(name)
        try:
            cert.load()
        except self._iam_client.exceptions.NoSuchEntityException:
            pass
        else:
            raise CertificateExistsException('A certificate named %s already exists in IAM.' % name)

        if not dry_run:
            response = self._iam_client.upload_server_certificate(
                Path=iam_path,
                ServerCertificateName=name,
                CertificateBody=cert_data,
                PrivateKey=private_key_data,
                CertificateChain=chain_data
            )

            certificate = Certificate.from_metadata(response['ServerCertificateMetadata'])

            # Print details about the upload
            print('Certificate Uploaded Successfully:')
            print('\tCertificate ID: %s' % certificate.id)
            print('\tCertificate Name: %s' % certificate.name)
            print('\tARN: %s' % certificate.arn)
            print('\tPath: %s' % certificate.path)
            print('\tUpload Date: %s' % certificate.upload_date)
            print('\tExpiration Date: %s' % certificate.expiration_date)

            return certificate.arn
        else:
            print('%sWould upload certificate to IAM:' % '[DRY RUN] ' if dry_run else '')
            print('\tName: %s' % name)
            print('\tPath: %s' % iam_path)
            print('\tCertificate File Path (local system): %s' % cert_data)
            print('\tPrivate Key File Path (local system): %s' % private_key_data)
            print('\tCertificate Chain File Path (local system): %s' % chain_data)



class CertificateExistsException(Exception):
    pass


class CertificateExpiredException(Exception):
    pass


class CertificateScanFailedException(Exception):
    pass


class DuplicatePrivateKeyException(Exception):
    pass


class DuplicateCertificateException(Exception):
    pass


class ListenerNotFoundException(Exception):
    pass


class InvalidProtocolException(Exception):
    pass


class LoadBalancerNotFoundException(Exception):
    pass


class MultipleCertificatesException(Exception):
    pass



class Certificate:
    """Certificate object that stores metadata about an IAM server certificate"""

    def __init__(self, name, id, path, arn, upload_date, expiration_date):
        self.name = name
        self.id = id
        self.path = path
        self.arn = arn
        self.upload_date = upload_date
        self.expiration_date = expiration_date

    @classmethod
    def from_metadata(cls, data):
        return cls(
            name=data['ServerCertificateName'],
            id=data['ServerCertificateId'],
            path=data['Path'],
            arn=data['Arn'],
            upload_date=data['UploadDate'],
            expiration_date=data['Expiration']
        )


class CertificateChain:
    """Object to store the tree of a certificate chain."""

    parent = None
    child = None

    def __init__(self, cert, path):
        """
        Initialize the certificate chain

        :param OpenSSL.crypto.X509 cert: X509 certificate
        :param str path: File path
        """
        self.cert = cert
        self.path = path



from setuptools import setup, find_packages

with open('README.md', 'r') as f:
    long_description = f.read()

setup(
    name='cert-uploader',
    description='Upload ACM/IAM Server Certificates to AWS and apply them to ELBs',
    long_description=long_description,
    long_description_content_type='text/markdown',
    keywords='certificate cert uploader tools aws iam acm',
    url='https://github.com/GESkunkworks/cert-uploader',
    use_scm_version=True,
    author='Michael Palmer',
    author_email='github@michaeldpalmer.com',
    packages=find_packages(),
    setup_requires=['setuptools_scm~=2.0.0'],
    install_requires=[
        'boto3~=1.7.16',
        'cryptography~=2.2.2',
        'pyOpenSSL~=17.5.0',
        'scandir~=1.7;python_version<"3.0"',
        'six~=1.11.0'
    ],
    entry_points={
        'console_scripts': [
            'cert-uploader = cert_uploader.cli:main',
        ],
    },
    classifiers=[
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.6',
        'Development Status :: 4 - Beta',
        'Topic :: Utilities'
    ]
)
